// Database.cpp
//
// Zugriff auf Adabas
//

#include <WINDOWS.H>
#include <sql.h>
#include "database.h"
#include "collection.h"
#include "util.h"
#include "strcoll.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// Puffer fuer interne Verwendung
char	buffer[1024];

// tDatabase

tDatabase::tDatabase(char *n, char *u, char *p)
{
	uid = new char[strlen(u)+1];
	pwd = new char[strlen(p)+1];
	name = new char[strlen(n)+1];

	strcpy(uid, u);
	strcpy(pwd, p);
	strcpy(name, n);

	working = 0;
}

tDatabase::~tDatabase()
{
	close();
	delete uid;
	delete pwd;
	delete name;
}

void tDatabase::open()
{
	RETCODE rt;

	SQLAllocEnv(&henv);
	SQLAllocConnect(henv, &hdbc);
	sprintf(buffer, "Connect to datasource %s as user %s", name, uid);
	logmsg(buffer, 1);
	rt = SQLConnect(hdbc, (UCHAR*)name, SQL_NTS, (UCHAR*)uid, SQL_NTS, (UCHAR*)pwd, SQL_NTS);
	if(rt!=SQL_SUCCESS && rt!=SQL_SUCCESS_WITH_INFO)
	{
		print_err(0);
		return;
	}
	working = 1;
	logmsg("database in working state.", 1);
}

void tDatabase::close()
{
	if(working)
	{
		sprintf(buffer, "Disconnect from datasource %s as user %s", name, uid);
		working=0;
		logmsg("database closed.", 1);
		logmsg(buffer, 1);
		SQLDisconnect(hdbc);
		SQLFreeConnect(hdbc);
		SQLFreeEnv(henv);
	}
}

void tDatabase::print_err(HSTMT hstmt)
{
	UCHAR	sqlstate[10];
	SDWORD	sqlcode;
	UCHAR	error_txt[256];
	SWORD	len_error_txt=255;
	SWORD	used_error_txt;
	
	SQLError(SQL_NULL_HENV, hdbc, hstmt, sqlstate, &sqlcode, error_txt, len_error_txt, &used_error_txt);

	sprintf(buffer, "sqlstate: %s", sqlstate); logmsg(buffer, 2);
	sprintf(buffer, "sqlcode: %ld '%s'", sqlcode, error_txt); logmsg(buffer, 2);
}

void tDatabase::getHSTMT(HSTMT &hstmt)
{
	if(SQLAllocStmt(hdbc, &hstmt)!=SQL_SUCCESS)
		print_err(0);
}

void tDatabase::freeHSTMT(HSTMT &hstmt)
{
	SQLFreeStmt(hstmt, SQL_DROP);
}

void tDatabase::ExecSQL ( char *s )
{
	tDataSource *ds = new tDataSource(this);

	ds->ExecSQL(s);
	
	delete ds;
}

void tDatabase::CommitWork()
{
	SQLTransact(henv, hdbc, SQL_COMMIT);
}

// tDataSource

tDataSource::tDataSource( char *s, tDatabase *d )
{
	sql = new char[strlen(s)+1];
	database = d;
	hstmt = 0;
	rows = NULL;

	strcpy(sql, s);

	data = NULL;
	rows = NULL;
} 

tDataSource::tDataSource(tDatabase *d)
{
	database=d;
	hstmt=0;
	rows=NULL;
	sql=NULL;

	data = NULL;
	rows = NULL;
}

tDataSource::~tDataSource()
{
	reset();
	delete sql;
	delete data;
	delete rows;
}

void tDataSource::ExecSQL( char *s )
{
	if(database==NULL)
	{
		logmsg("ERROR: datasource -> no database assigned", 2);
		return;
	}
	if(sql!=NULL)
		delete sql;
	sql = new char[strlen(s)+1];
	strcpy(sql, s);
	database->getHSTMT(hstmt);
	if(SQLPrepare(hstmt, (UCHAR*)sql, SQL_NTS)!=SQL_SUCCESS)
	{
		database->print_err(hstmt);
	}
	data = new tCollection(15, 15);
	rows = new tStrColl(15, 15);
	active = 1;
	RunSQL();
}

void tDataSource::reset()
{
	if(active)
	{
		active=0;
		logmsg("Datasource reset.", 0);
		database->freeHSTMT(hstmt);
		hstmt = 0;
		nresultcols = rowcount = position = 0;
		delete rows;
		delete data;
		rows = NULL;
		data = NULL;
	}
	else
	{
		logmsg("Warning: Reset of inactive dataset.", 1);
	}
}

void tDataSource::RunSQL()
{
	UCHAR colname[32];
	SWORD colnamelen;
	SWORD coltype;
	UDWORD collen[MAXCOLS];
	SWORD scale;
	SWORD nullable;
	SDWORD outlen[MAXCOLS];
	UCHAR *tempo[MAXCOLS];
	RETCODE rc;
	long cnt;
	char *buf;
	tStrColl *tmp;

	cnt = 0;

	if(SQLExecute(hstmt)!=SQL_SUCCESS)
	{
		database->print_err(hstmt);
		sprintf(buffer, "Query FAILED: \"%s\".", sql);
		logmsg(buffer, 2);
		return;
	}
	sprintf(buffer, "Query \"%s\" successfully performed.", sql);
	logmsg(buffer, 0);

	// determine the result type
	// and get the data...

	SQLNumResultCols(hstmt, &nresultcols);

	if(nresultcols==0) {
		// only affecting query, no results :-((
		SQLRowCount(hstmt, &rowcount);
		if(rowcount>0) {
			sprintf(buffer, "%ld rows affected by query \"%s\".", rowcount, sql); logmsg(buffer, 1);
		}
		database->CommitWork();
	}
	else
	{
		// get the data :-))
		sprintf(buffer, "%ld cols returned by query \"%s\".", nresultcols, sql); logmsg(buffer, 1);
		for(int i=0; i<nresultcols; i++)
		{
			SQLDescribeCol(hstmt, (UWORD)(i+1), colname, (SWORD)sizeof(colname), &colnamelen, &coltype, &collen[i], &scale, &nullable);
			tempo[i] = (UCHAR *)malloc(collen[i]+5);
			SQLBindCol(hstmt, (UWORD)(i+1), SQL_C_CHAR, tempo[i], collen[i]+1, &outlen[i]);
			buf = new char[colnamelen+1];
			strcpy(buf, colname);
			//printf("blubb...\n");
			rows->Add(buf);
			buf = NULL;
		}

		//printf("get data...\n");

		while (TRUE)
		{
			rc = SQLFetch(hstmt);
			if(rc==SQL_SUCCESS || rc==SQL_SUCCESS_WITH_INFO) {
				tmp = new tStrColl();
				data->Add(tmp);
				cnt++;
				for(int i=0; i<nresultcols; i++) {
					if(outlen[i]==SQL_NULL_DATA) {
						strcpy((char*)tempo[i], "");
					}
					//printf("alloc buffer %d ...\n ", i);
					buf = new char[outlen[i]+2];
					//printf("copy buffer...\n  ");
					strcpy(buf, (char*)tempo[i]);
					//printf("add buffer (len: %d): %s\n", strlen(buf), buf);
					tmp->Add(buf);
					//printf("    ...done\n");
					buf=NULL;
				}
				tmp = NULL;
			}
			else
			{	
				break;
			}
		}
		for(int i=0; i<nresultcols; i++)
			free(tempo[i]);
		sprintf(buffer, "%d rows returned by query \"%s\".", cnt, sql); logmsg(buffer, 1);
		rowcount = cnt;
		//printf("data for %d rows retrieved...\n", cnt);
	}
}

tStrColl *tDataSource::getRow(int i)
{
	return (tStrColl*)(data->Get(i));
}

void tDataSource::close()
{
	if(database==NULL)
	{
	        logmsg("ERROR: datasource -> no database assigned", 2);
	        return;
	}
	if(!hstmt)
	{
		logmsg("Warning: non existent hstmt closed.", 1);
		return;
	}
	SQLFreeStmt(hstmt, SQL_DROP);
	hstmt = 0;
}
